module Minicontext
  class Context
    KEYWORD = %i(before after)
    COMPOSE = -> (f, g) { -> (env) { f.(g.(env)) } }
    RELAY   = -> (env) { env }

    def initialize(parent = nil)
      @vars   = {}
      @parent = parent
    end

    def context(name, &block)
      raise AssignError, name if KEYWORD.include?(name)
      raise AssignError, name if (val = @vars[name]) && Proc === val

      val ? val.action(&block) : define_context(name, &block)
    end

    def task(name, &block)
      self[name] = -> (env) { env.(self, &block) }
    end

    def let(name, &block)
      raise AssignError, name if KEYWORD.include?(name)

      self[name] = -> (env) do
        env.memoized.fetch([self, name]) do
          env.memoized[[self, name]] = env.(self, &block)
        end
      end
    end

    %i(before after).each do |name|
      define_method(name) {|&block| task name, &block}
    end

    def action(&block)
      instance_eval &block
    end

    def load(filename)
      instance_eval open(filename) {|f| f.read}, filename
    end

    def call(name)
      Proc === (val = @vars[name]) ? apply(name) : val
    end

    def lookup(name)
      Proc === (val = @vars[name]) ? val : (@parent && @parent.lookup(name))
    end

    def [](name)
      router[name]
    end

    def []=(name, val)
      raise AssignError, name if @vars[name]

      @vars[name] = val
    end

    private

    def define_context(name, &block)
      self.class.new(self).tap do |ctx|
        self[name] = ctx
        ctx.action &block
      end
    end

    def apply(name)
      compose(@vars[name])
    end

    def compose(f)
      g = COMPOSE.(@vars[:after] || RELAY, COMPOSE.(f, @vars[:before] || RELAY))

      @parent ? @parent.send(:compose, g) : g
    end

    def router
      Router::Maybe.new(self)
    end
  end
end
